package sugui.day07;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import sugui.day07.Hand.HandType;

public class HandWithJokerComparator implements Comparator<Hand> {

    private HandType getHandType(Hand hand) {
        Map<String, Integer> cardCount = new HashMap<>();
        int jokerCount = 0;
        for (Card c : List.of(hand.c1(), hand.c2(), hand.c3(), hand.c4(), hand.c5())) {
            if ("J".equals(c.label())) {
                jokerCount++;
            } else if (cardCount.containsKey(c.label())) {
                cardCount.compute(c.label(), (k, v) -> v + 1);
            } else {
                cardCount.put(c.label(), 1);
            }
        }
        List<Integer> counts = new ArrayList<>(cardCount.values());
        counts.sort(Comparator.reverseOrder());
        if (counts.size() == 0) {
            // Hand of full Jokers
            return HandType.FIVE_OF_A_KIND;
        } else {
            int fstCount = counts.get(0) + jokerCount;
            int sndCount = counts.size() == 1 ? 0 : counts.get(1);
            if (fstCount == 5)
                return HandType.FIVE_OF_A_KIND;
            else if (fstCount == 4)
                return HandType.FOUR_OF_A_KIND;
            else if (fstCount == 3 && sndCount == 2)
                return HandType.FULL_HOUSE;
            else if (fstCount == 3 && sndCount == 1)
                return HandType.THREE_OF_A_KIND;
            else if (fstCount == 2 && sndCount == 2)
                return HandType.TWO_PAIR;
            else if (fstCount == 2 && sndCount == 1)
                return HandType.ONE_PAIR;
            else
                return HandType.HIGH_CARD;
        }
    }

    @Override
    public int compare(Hand h1, Hand h2) {
        HandType h1HandType = getHandType(h1);
        HandType h2HandType = getHandType(h2);

        if (h1HandType.equals(h2HandType)) {
            var cardComparator = new CardWithJokerComparator();
            int cmp1 = cardComparator.compare(h1.c1(), h2.c1());
            int cmp2 = cardComparator.compare(h1.c2(), h2.c2());
            int cmp3 = cardComparator.compare(h1.c3(), h2.c3());
            int cmp4 = cardComparator.compare(h1.c4(), h2.c4());
            int cmp5 = cardComparator.compare(h1.c5(), h2.c5());
            return Stream.of(cmp1, cmp2, cmp3, cmp4, cmp5)
                    .dropWhile(cmp -> cmp == 0)
                    .findFirst()
                    .orElse(0);
        } else {
            return h1HandType.compareTo(h2HandType);
        }
    }

}
